using System.Net;
using System.Reflection;
using System.Security.Cryptography;
using System.Text;
using Autofac;
using Autofac.Core.Activators.Reflection;
using FitConnect.Encryption;
using FitConnect.Interfaces.Subscriber;
using FitConnect.Models;
using FitConnect.Models.v1.Api;
using FitConnect.Services.Models;
using FitConnect.Services.Models.v1.Submission;
using IdentityModel;
using Microsoft.AspNetCore.Http;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using NJsonSchema;
using NJsonSchema.Annotations;
using NJsonSchema.Validation;
using Metadata = FitConnect.Models.Api.Metadata.Metadata;

namespace FitConnect;

/// <summary>
///     Fluent API for the FitConnect.Subscriber
/// </summary>
public class Subscriber : FitConnectClient,
    ISubscriber,
    ISubscriberWithSubmission {
    public Submission? Submission { get; private set; }

    public Subscriber(FitConnectEnvironment environment,
        string clientId, string clientSecret,
        string privateKeyDecryption,
        string privateKeySigning,
        IContainer container) : base(environment, clientId, clientSecret,
        container,
        privateKeyDecryption, privateKeySigning) {
        Encryption = new FitEncryption(privateKeyDecryption, privateKeySigning,
            container.Resolve<ILogger>());
    }

    public Subscriber(FitConnectEnvironment environment,
        string clientId, string clientSecret,
        string privateKeyDecryption,
        string privateKeySigning,
        ILogger? logger = null) : base(environment, clientId, clientSecret, logger,
        privateKeyDecryption, privateKeySigning) {
        Encryption = new FitEncryption(privateKeyDecryption, privateKeySigning, logger);
    }


    /// <summary>
    ///     Receives the available submissions from the server
    /// </summary>
    /// <param name="destinationId"></param>
    /// <param name="skip"></param>
    /// <param name="take"></param>
    /// <returns></returns>
    public async Task<IEnumerable<SubmissionForPickupDto>> GetAvailableSubmissionsAsync(
        string? destinationId = null,
        int skip = 0,
        int take = 100) {
        var submissionsResult = await SubmissionService.ListSubmissionsAsync(destinationId, 0, 100);

        // Creating a dictionary of destinationId to submissionIds from the REST API result
        return submissionsResult.Submissions ?? new List<SubmissionForPickupDto>();
    }


    /// <summary>
    ///     Receives a specific submission from the server.
    /// </summary>
    /// <param name="submissionId"></param>
    /// <param name="skipSchemaTest"></param>
    /// <returns></returns>
    public async Task<ISubscriberWithSubmission> RequestSubmissionAsync(string submissionId,
        bool skipSchemaTest = false) {
        var submission = (Submission)await SubmissionService.GetSubmissionAsync(submissionId);
        var (metaDataString, _, metaHash) = Encryption.Decrypt(submission.EncryptedMetadata!);

        if (!skipSchemaTest) {
            var valid = await JsonHelper.VerifyMetadata(metaDataString);
            if (!valid) {
                Logger?.LogWarning("Invalid metadata: {MetaData}", metaDataString);

                throw new Exception($"Metadata validation failed");
            }
        }

        submission.Metadata = JsonConvert.DeserializeObject<Metadata>(metaDataString);

        if (submission.EncryptedData != null) {
            var (dataString, _, dataHash) = Encryption.Decrypt(submission.EncryptedData);
            submission.Data = dataString;

            if (submission.Metadata?.ContentStructure.Data.Hash.Content !=
                FitEncryption.CalculateHash(dataString)) {
                Logger?.LogWarning("Data hash mismatch: {DataHash} != {CalculatedHash}",
                    submission.Metadata?.ContentStructure.Data.Hash.Content,
                    FitEncryption.CalculateHash(dataString));
                throw new Exception("Data hash mismatch");
            }
        }

        Submission = submission;
        return this;
    }


    /// <summary>
    ///     Reading attachments for a submission.
    /// </summary>
    /// <returns></returns>
    public async Task<IEnumerable<Attachment>> GetAttachmentsAsync() {
        if (Submission?.Id == null || Submission?.Metadata == null)
            throw new Exception("No submission available");

        var attachments = new List<Attachment>();
        foreach (var id in Submission!.AttachmentIds) {
            var encryptedAttachment = await SubmissionService.GetAttachmentAsync(Submission.Id, id);
            var (_, content, hash) = Encryption.Decrypt(encryptedAttachment);
            var attachmentMeta =
                Submission.Metadata.ContentStructure.Attachments.First(a => a.AttachmentId == id);

            attachments.Add(new Attachment(attachmentMeta, content,
                encryptedAttachment.Split('.').Last()) {
                Description = attachmentMeta.Description,
                Purpose = attachmentMeta.Purpose,
                MimeType = attachmentMeta.MimeType
            });
        }

        Submission.Attachments = attachments;
        return attachments;
    }

    public async Task AcceptSubmissionAsync() =>
        await CompleteSubmission(Submission!, FinishSubmissionStatus.Accepted);

    public async Task RejectSubmissionAsync(params Problems[] problems) =>
        await CompleteSubmission(Submission!, FinishSubmissionStatus.Rejected, problems);


    // public void ForwardSubmission() {
    //     CompleteSubmission(Submission!, FinishSubmissionStatus.Forwarded);
    // }

    public async Task CompleteSubmission(FinishSubmissionStatus status) =>
        await CompleteSubmission(Submission!, status);

    public bool VerifyStatus(AcceptanceStatus acceptanceStatus) {
        if (Submission == null)
            throw new NullReferenceException("Submission is null");
        var result = true;
        result &= acceptanceStatus.Metadata == Submission.MetaAuthentication;
        result &= acceptanceStatus.Data == Submission.DataAuthentication;
        if (Submission.Attachments != null)
            foreach (var attachment in Submission.Attachments) {
                result &= acceptanceStatus.Attachments[attachment.Id] ==
                          attachment.AttachmentAuthentication;
            }

        return result;
    }


    /// <summary>
    /// Accept or reject submission
    /// </summary>
    private async Task CompleteSubmission(Submission submission,
        FinishSubmissionStatus status, Problems[]? problems = null) {
        if (submission.Id == null || submission.CaseId == null ||
            submission.DestinationId == null)
            throw new ArgumentException("Submission does not contain all required fields");

        if (status != FinishSubmissionStatus.Rejected && problems != null)
            throw new ArgumentException("Problems can only be set for rejected submissions");

        var token = status switch {
            FinishSubmissionStatus.Rejected =>
                await Encryption.CreateRejectSecurityEventToken(submission.Id,
                    submission.CaseId,
                    submission.DestinationId, problems),
            FinishSubmissionStatus.Accepted => await Encryption.CreateAcceptSecurityEventToken(
                submission),
            _ => throw new ArgumentOutOfRangeException(nameof(status), status, null)
        };

        Logger?.LogDebug("Token to accept submission: {Token}", token);
        var result = await CasesService.FinishSubmission(submission.CaseId, token);
        Logger?.LogInformation("Submission completed {Status}", result);
    }

    public static string VerifyCallback(string callbackSecret,
        long timestamp, string body) {
        if (timestamp < DateTime.Now.AddMinutes(SecuritySpecification.MaxCallbackAge * -1)
                .ToEpochTime())
            throw new ArgumentException("Request is too old");
        var hmac = SecuritySpecification.CalculateCallbackHmac(callbackSecret, timestamp, body);
        return Convert.ToHexString(hmac).ToLower();
    }


    public static bool VerifyCallback(string callbackSecret, HttpRequest request) {
        if (!request.Headers.ContainsKey("callback-timestamp"))
            throw new ArgumentException("Missing callback-timestamp header");

        var timeStampString = request.Headers["callback-timestamp"].ToString();
        if (!long.TryParse(timeStampString, out var timestamp)) {
            throw new ArgumentException("Invalid callback-timestamp header");
        }


        var authentication = request.Headers["callback-authentication"];
        using var requestStream = request.Body;
        var content = new StreamReader(requestStream).ReadToEnd().Trim();

        var result = VerifyCallback(callbackSecret, timestamp, content);
        if (result != authentication)
            throw new ArgumentException("Verified request does not match authentication");
        return true;
    }
}

public enum FinishSubmissionStatus {
    Accepted,

    Rejected
    // Forwarded
}
